#include "parser.h"

#include <stdio.h>
#include <stdlib.h>

#include "model.h"

/* prototypes */
struct expression *primary (struct parser *p);
struct expression *unary (struct parser *p);
struct expression *factor (struct parser *p);
struct expression *term (struct parser *p);
struct expression *expression (struct parser *p);

/* util functions */
/* static struct token * */
/* advance (struct parser *p) */
/* { */
/*	if (!p) */
/*		return NULL; */

/*	if (p->curr >= vector_len (p->tokens)) */
/*		return NULL; */

/*	struct token *tok = (struct token *)vector_at (p->curr, p->tokens); */
/*	p->curr++; */
/*	return tok; */
/* } */

/* static struct token * */
/* peek (struct parser *p) */
/* { */
/*	if (!p) */
/*		return NULL; */

/*	return (struct token *)vector_at (p->curr, p->tokens); */
/* } */

/* static _Bool */
/* expect (enum token_type type, struct parser *p) */
/* { */
/*	if (!p) */
/*		return 0; */

/*	struct token *tok = (struct token *)vector_at (p->curr, p->tokens); */
/*	return tok->type == type; */
/* } */

static struct token *
prev_tok (struct parser *p)
{
	if (!p)
		return NULL;

	if (p->curr <= 0)
		return NULL;

	return (struct token *)vector_at (p->curr - 1, p->tokens);
}

static struct token *
match (enum token_type type, struct parser *p)
{
	if (!p)
		return NULL;

	struct token *tok = (struct token *)vector_at (p->curr, p->tokens);
	if (!tok || tok->type != type)
		return NULL;

	p->curr++;
	return tok;
}

/* recursively descending functions */
struct expression *
primary (struct parser *p)
{
	if (!p)
		return NULL;

	struct token *tok = NULL;
	if ((tok = match (TOK_INTEGER, p)))
		return expression_create_integer (atoi (buffer_get (&tok->lexeme)));
	if ((tok = match (TOK_FLOAT, p)))
		return expression_create_floating (atof (buffer_get (&tok->lexeme)));
	if ((tok = match (TOK_LPAREN, p)))
		{
			struct expression *expr = expression (p);
			if (!(tok = match (TOK_RPAREN, p)))
				{
					fprintf (stderr, "Error: `)' expected at line %d.\n", tok->line);
					exit (1);
				}
			return expression_create_grouping (expr);
		}
	return NULL; /* unmatched */
}

struct expression *
unary (struct parser *p)
{
	if (!p)
		return NULL;

	if (match (TOK_NOT, p) || match (TOK_MINUS, p) || match (TOK_PLUS, p))
		{
			struct token *op = prev_tok (p);
			struct expression *right = unary (p);
			return expression_create_unop (op, right);
		}

	return primary (p);
}

struct expression *
factor (struct parser *p)
{
	if (!p)
		return NULL;

	return unary (p);
}

struct expression *
term (struct parser *p)
{
	if (!p)
		return NULL;

	struct expression *expr = factor (p);
	while (match (TOK_STAR, p) || match (TOK_SLASH, p))
		{
			struct token *op = prev_tok (p);
			struct expression *right = factor (p);
			expr = expression_create_binop (op, expr, right);
		}

	return expr;
}

struct expression *
expression (struct parser *p)
{
	if (!p)
		return NULL;

	struct expression *expr = term (p);
	while (match (TOK_PLUS, p) || match (TOK_MINUS, p))
		{
			struct token *op = prev_tok (p);
			struct expression *right = term (p);
			expr = expression_create_binop (op, expr, right);
		}

	return expr;
}

struct parser
parser_create (struct vector *tokens)
{
	struct parser parser = { 0 };
	parser.tokens = tokens;
	parser.ast = vector_create ();
	return parser;
}

void
parser_parse (struct parser *p)
{
	if (!p)
		return;

	vector_push_back (expression (p), &p->ast);
}

void
expr_print (struct expression *expr)
{
	switch (expr->type)
		{
		case EXPR_INT:
			printf ("Integer[%d]", expr->integer);
			break;
		case EXPR_FLOAT:
			printf ("Float[%f]", expr->floating);
			break;
		case EXPR_BINOP:
			printf ("BinOp('%s', ", buffer_get (&expr->binop.op->lexeme));
			expr_print (expr->binop.left);
			printf (", ");
			expr_print (expr->binop.right);
			printf (")");
			break;
		case EXPR_UNOP:
			printf ("UnOp('%s', ", buffer_get (&expr->unop.op->lexeme));
			expr_print (expr->unop.operand);
			printf (")");
			break;
		case EXPR_GROUP:
			printf ("Grouping(");
			expr_print (expr->grouping.value);
			printf (")");
			break;
		default:
			fprintf (stderr, "Unknown expression type: %d\n", expr->type);
			break;
		}
}

void
parser_print (struct parser *p)
{
	puts ("Parsed Ast:");
	for (unsigned int i = 0; i < vector_len (&p->ast); i++)
		{
			struct expression *expr = (struct expression *)vector_at (i, &p->ast);
			expr_print (expr);
			printf ("\n");
		}
}
